/*
 *  This file is part of BGX, the Bayesian Gene eXpression program.
 *  Copyright (c) 2003-2004  Graeme Ambler <graeme@ambler.me.uk>
 *                2005-2008  Ernest Turro <ernest.turro@ic.ac.uk>
 *
 *  BGX is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, version 2, as
 *  published by the Free Software Foundation.
 *
 *  BGX is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef _BGX_UPDATES_HH
#define _BGX_UPDATES_HH

#include "rwm.hh"

#ifdef _OPENMP
  #include <omp.h>
#endif // _OPENMP
#include <valarray>

using namespace std; 
typedef valarray<double> array;
typedef valarray<valarray<double> > array2d;

#define SAFE_EXP(a) exp(max(-500.0,min(0.0,a)))

inline double pnorm(double x){ return .5+.5*erf(M_SQRT1_2*x); }

class S_T{
private:
  array2d& PM;
  array2d& mu;
  array2d& sigma;
  array2d& gamma;
  array2d& delta;
  int* samplesets;
  int* categories;
  vector<int> *pm2ps;
  vector<int> *ps2pm;
  vector<int> replicate2condition;
public:
  S_T(array2d& PM_, array2d& mu_, 
      array2d& sigma_, array2d& gamma_, array2d& delta_,
      int* samplesets_, int * categories_, vector<int> *pm2ps_, vector<int> *ps2pm_)
    : PM(PM_),mu(mu_), sigma(sigma_), 
      gamma(gamma_), delta(delta_), samplesets(samplesets_), categories(categories_) {
        pm2ps=pm2ps_;
        ps2pm=ps2pm_;
        for(int c=0; c < (int)mu.size(); c++) {
          for(int r=0; r < samplesets[c]; r++) {
            replicate2condition.push_back(c);
          }
        }
      }

  double operator()(array2d& S, double S2, int r, int j)
  {
    if(S2>0 && S2 < PM[r][j]) {
      double x=log(S[r][j]);
      double x2=log(S2);

      double logsumexpsigma =0, logsumexpmu = 0;
      for(unsigned int g=0; g < pm2ps[j].size(); g++) {
        logsumexpsigma += exp(sigma[replicate2condition[r]][pm2ps[j][g]]); // logsumexpsigma is log sum of variances
        logsumexpmu += exp(mu[replicate2condition[r]][pm2ps[j][g]]);
      }
      logsumexpsigma = log(logsumexpsigma);
      logsumexpmu= log(logsumexpmu);

      double H = PM[r][j] - S[r][j];
      double H2 = PM[r][j] - S2;

      return SAFE_EXP(x - x2 + 
              .5/logsumexpsigma * (
                (x-logsumexpmu)*(x-logsumexpmu) - 
                (x2-logsumexpmu)*(x2-logsumexpmu)) + 
              log(H/H2) + 
              .5/delta[r][categories[j]] *
                ( pow(log(H) - gamma[r][categories[j]],2) -
                  pow(log(H2) - gamma[r][categories[j]],2)));

    } else return 0;
  }
};

class Mu_T{
private:
  array2d& S;
  array2d& sigma;
  int* samplesets;
  int *offsets;
  vector<int> *pm2ps;
  vector<int> *ps2pm;
public:
  Mu_T(array2d& S_, array2d& sigma_, int* samplesets_, vector<int> *pm2ps_, vector<int> *ps2pm_)
    : S(S_), sigma(sigma_), samplesets(samplesets_) {
    pm2ps=pm2ps_;
    ps2pm=ps2pm_;
    offsets = new int[sigma.size()]; 
    offsets[0] = 0;
    for(unsigned int k=1; k < sigma.size(); k++) {
      offsets[k] = offsets[k-1] + samplesets[k-1];  
    }
  }

  ~Mu_T() { delete offsets; }

  double operator()(array2d& mu, double mu2, int k, int i)
  {

    if(mu2> 0 && mu2<15) {
      double *logsumexpmu = new double[ps2pm[i].size()];
      double *logsumexpmu2 = new double[ps2pm[i].size()];
      double *logsumexpsigma = new double[ps2pm[i].size()]; // logsumexpsigma is log sum of variances
      for(unsigned int p=0; p < ps2pm[i].size(); p++) {
        logsumexpmu[p]=logsumexpmu2[p]=logsumexpsigma[p]=0; 
        for(unsigned int g=0; g < pm2ps[ps2pm[i][p]].size(); g++) {
          logsumexpmu[p] += exp(mu[k][pm2ps[ps2pm[i][p]][g]]);
          logsumexpsigma[p] += exp(sigma[k][pm2ps[ps2pm[i][p]][g]]);
        }
        logsumexpmu2[p] = logsumexpmu[p] - exp(mu[k][i]) + exp(mu2);

        logsumexpmu[p] = log(logsumexpmu[p]);
        logsumexpmu2[p] = log(logsumexpmu2[p]);
        logsumexpsigma[p] = log(logsumexpsigma[p]);
      }

      double sum=0;
      for(int l=0; l<samplesets[k]; ++l) {
        for(unsigned int p=0; p < ps2pm[i].size(); p++) {
          double x=log(S[offsets[k]+l][ps2pm[i][p]]);
          sum+= (0.5/logsumexpsigma[p]) * 
            ((x-logsumexpmu[p])*(x-logsumexpmu[p]) - (x - logsumexpmu2[p])*(x -logsumexpmu2[p]));
        }
      }
      delete logsumexpmu; delete logsumexpmu2; delete logsumexpsigma;
      return SAFE_EXP(sum);
    } else return 0;
  }
};

class Sigma_T{
private:
  array2d& S;
  array2d& mu;
  array& a;
  array& b;
  int* samplesets;
  vector<int> *pm2ps;
  vector<int> *ps2pm;
  int* offsets;
public:
  Sigma_T(array2d& S_, array2d& mu_, array& a_, array& b_, int* samplesets_, vector<int> *pm2ps_, vector<int> *ps2pm_)
    : S(S_), mu(mu_), a(a_), b(b_),samplesets(samplesets_){
    pm2ps=pm2ps_;
    ps2pm=ps2pm_;
    offsets = new int[mu.size()];
    offsets[0] = 0;
    for(unsigned int k=1; k < mu.size(); k++) {
      offsets[k] = offsets[k-1] + samplesets[k-1];
    }
  }

  ~Sigma_T() { delete offsets; }

  double operator()(array2d& sigma, double sigma2, int k, int i)
  {
    if(sigma2>0) {
      double *logsumexpmu = new double[ps2pm[i].size()];
      double *logsumexpsigma = new double[ps2pm[i].size()]; // log sum of variances
      double *logsumexpsigma2 = new double[ps2pm[i].size()]; // log sum of variances
      for(unsigned int p=0; p < ps2pm[i].size(); p++) {
        logsumexpmu[p]=logsumexpsigma[p]=logsumexpsigma2[p]=0;
        for(unsigned int g=0; g < pm2ps[ps2pm[i][p]].size(); g++) {
          logsumexpmu[p] += exp(mu[k][pm2ps[ps2pm[i][p]][g]]);
          logsumexpsigma[p] += exp(sigma[k][pm2ps[ps2pm[i][p]][g]]);
        }
        logsumexpsigma2[p] = log(logsumexpsigma[p] - exp(sigma[k][i]) + exp(sigma2));
        logsumexpmu[p] = log(logsumexpmu[p]);
        logsumexpsigma[p] = log(logsumexpsigma[p]);
      }

      double sum=0;  
      for(int l=0; l<samplesets[k]; ++l) {
        for(unsigned int p=0; p < ps2pm[i].size(); p++) {
          double x=log(S[offsets[k]+l][ps2pm[i][p]]);
          sum+= 0.5*(log(logsumexpsigma[p]/logsumexpsigma2[p]) +          
            ((1.0/logsumexpsigma[p]) - (1.0/logsumexpsigma2[p]))*(x - logsumexpmu[p])*(x-logsumexpmu[p])); 
        }
      }
      delete logsumexpmu;
      delete logsumexpsigma;
      delete logsumexpsigma2;
      
      return SAFE_EXP(sum
              -0.5/b[k]*(pow(log(sigma2)-a[k], 2) -
                pow(log(sigma[k][i])-a[k],2)) +log(sigma[k][i]/sigma2));
      } else return 0;
  }
};

/* Not in use, fixed by Emp Bayes like procedure
class A_T{
private:
  array& a;
  array2d& sigma;
  array& b;
  double p_mu;
  double p_tau;
  Random* rand;
  int genes;
  int conditions;
public:
  A_T(array& a_, array2d& sigma_, array& b_,
      double p_mu_, double p_tau_, Random* rand_)
    : a(a_), sigma(sigma_), b(b_),
      p_mu(p_mu_), p_tau(p_tau_), rand(rand_), genes(sigma[0].size()), conditions(sigma.size()) {}

  void Update()
  {
    for(int j=0; j<conditions; ++j)
      {
  double sum=0;
  for(int i=0; i<genes; ++i)
    {
      sum += log(sigma[j][i]);
    }
  a[j]=rand->Normal( (p_tau*p_mu+sum/b[j])/(p_tau+genes/b[j]),
         sqrt(1.0/(p_tau+ genes/b[j]))) ;
      }
  }
};
*/

class B_T {
private:
  array& b;
  array2d& sigma;
  array& a;
  double alpha;
  double beta;
  Random* rand;
  int genes;
  int conditions;
public:
  B_T(array& b_, array2d& sigma_, array& a_,
      double alpha_, double beta_, Random* rand_)
    : b(b_), sigma(sigma_), a(a_),
      alpha(alpha_), beta(beta_), rand(rand_), genes(sigma[0].size()), conditions(sigma.size()) {}

  void Update() {
    for(int j=0; j<conditions; ++j) {
      double rss=0;
      #pragma omp parallel for schedule(static) reduction(+:rss) 
      for(int i=0; i<genes; ++i) {
        rss += pow((log(sigma[j][i])-a[j]), 2);
      }
      b[j]=1/rand->Gamma( alpha+genes*.5, beta+rss*.5 );
    }
  }
};

#endif // _BGX_UPDATES_HH
